;
; 
; 13-06-2008
;
; Filename: "ParallelPIC.s"
; The program is the "firmware" stored into the controler of each board.
; The Configuration Words must be: 
; CW1: x011 1x11 - 0111 1111  = 3F 7F (JTAG:dis, WDT:dis)
; CW2: 1111 1000 - 1100 1100  = F8 CC (Fast RC, dis ODC out, en IOLOCK)
;
; (The program assumes internal clock of 8MHZ. All delay calculations must be computed again !!!)
;******************************************************************************

               .equ __24FJ16GA002, 1
               .include "p24FJ16GA002.inc"

;..............................................................................
; Constants:
;..............................................................................
.equ	XTAL, 2			; The XTAL frequency in MHZ
.equ	VISI,  0x0784
.equ	PGC_HP,	15		; Half Period for PGC (15 for 32MHZ to generate 1usec) 
.equ	PGD_ID, 15		; PORTB number for current PGD 
;..............................................................................
;Global Declarations:
;..............................................................................

        .global __reset          ;The label for the first line of code. 

;..............................................................................
;Uninitialized variables in Near data memory (Lower 8Kb of RAM)
;..............................................................................

          .section .nbss, bss, near
InBuf:     		.space 256               ;256 bytes of space for variable "InBuf".
OutBuf:			.space 256				;256 bytes of space for variable "OutBuf"
OutBuf_index:	.space	2			;2 bytes (1 word) for indexing the OutBuf
VISIout:		.space	2			;2 bytes (1 word) for the data read from the slave's VISI.
BlockNumber:	.space	2			;2 bytes (1 word) number of block for Write Flash
BlockData:		.space	64*4		;space for 64 instructions (64*3 bytes required)
									;Data are written MSB to LSB, 3 bytes/instruction, 
									;LSAddress first!!!
CurrentMCU:		.space	2			;2 bytes (1 word) number of currently selected MCU
;..............................................................................
; Macro section
;..............................................................................

;
; Wait for 25 msec
;
; 25msec = 25 * 1msec loop
;
.macro wait_p7

		PUSH	W8
		NOP

		MOV		#25,W8
3:
		CALL	wait_p19
		NOP
		DEC		W8,W8
		BRA		NZ,3b


		POP	W8
.endm

.macro	PrintStringFlash		stringaddress
		PUSH	W0

		MOV	#tblpage(&stringaddress&),W0
		MOV	W0,_TBLPAG
		MOV	#tbloffset(&stringaddress&),W0

		CALL	UART_Print_String_flash

		POP		W0
.endm

.macro	PrintStringRAM		stringaddress
		PUSH	W0

		MOV	#&stringaddress&,W0

		CALL	UART_Print_String_RAM

		POP		W0
.endm

.macro	strcmp		RAMaddress  ,   FLASHaddress

		PUSH	W0
		PUSH	W1

		MOV		#&RAMaddress&,W0
		MOV		#tblpage(&FLASHaddress&),W1
		MOV		W1,_TBLPAG
		MOV		#tbloffset(&FLASHaddress&),W1

		CALL	stringcompare

		POP		W1
		POP		W0
.endm

.macro	ICSP_execute_code  addr
		MOV		#tblpage(&addr&),W1
		MOV		W1,_TBLPAG
		MOV		#tbloffset(&addr&),W3
		CALL	ICSP_send_code
.endm

.macro 	ICSP_send_opcode	Hi, Lo
		MOV		#0x00&Hi&,W2
		MOV		#0x&Lo&,W1
		CALL	ICSP_send_opcodeP
.endm

.macro	ICSP_set_MCLR_0
		BCLR	LATB,#6		;set MCLR=0		
.endm
.macro	ICSP_set_MCLR_1
		BSET	LATB,#6		;set MCLR=1
.endm
.macro	ICSP_set_PGC_0
		BCLR	LATB,#7		;set PGC=0		
.endm
.macro	ICSP_set_PGC_1
		BSET	LATB,#7		;set PGC=1
.endm

.macro	ICSP_set_PGD_0				;must be a one (1) command macro
		AND		W13,[W11],[W11]
.endm

.macro	ICSP_set_PGD_1				;must be a one (1) command macro
		IOR		W12,[W11],[W11]
.endm

;=======================

;..............................................................................
; Constant data Section in Program Memory
;..............................................................................
.text

prompt:			.ascii	">",<0x0>
CR:				.ascii	<0xD>,<0xA>,<0x0>
dash:			.ascii	" - ",<0x0>
idname: 		.ascii	"ParallelPIC V 1.0",<0xD>,<0xA>,<0x0>
block:			.ascii	"block",<0xD>,<0xA>,<0x0>
Unknown_cmd:		.ascii	"Unknown command",<0xD>,<0xA>,<0x0>

cmd_Version:		.ascii	"Version",<0x0>
cmd_Identify:		.ascii	"Identify",<0x0>
cmd_ReadCW:			.ascii	"ReadCW",<0x0>
cmd_ReadFlash:		.ascii	"ReadFlash",<0x0>
cmd_ReadFlashBin:	.ascii	"ReadFlashBin",<0x0>
cmd_MCLR0:			.ascii	"MCLR0",<0x0>
cmd_MCLR1:			.ascii	"MCLR1",<0x0>
cmd_Erase:			.ascii	"Erase",<0x0>
cmd_WriteCW:		.ascii	"WriteCW",<0x0>
cmd_WriteBlock:		.ascii	"WriteBlock",<0x01>,<0x00>
cmd_SelectMCU:		.ascii	"SelectMCU",<0x01>,<0x00>

				.palign 2
Command_set_Reset:
		NOP
		GOTO		0x200
		NOPR


Command_set_Read_ID_Hi:

		MOV			#0x00FF,W0
		MOV			W0,TBLPAG
		MOV			#0x0000,W6
		NOP
		NOP
		MOV			#0x0000,W7
		TBLRDL		[W6++],[W7]
		NOP
		NOP
		MOV			W0,0x0784
		NOP
		NOPR

Command_set_Read_ID_Lo:

		MOV			#0x0002,W6
		NOP
		NOP
		MOV			#0x0000,W7
		TBLRDL		[W6++],[W7]
		NOP
		NOP
		MOV			W0,0x0784
		NOP
		NOPR

Command_set_Read_CW_Hi:
		NOP
		GOTO		0x200

        MOV  		#0x0000,W0
        MOV  		W0,TBLPAG
        MOV  		#0x2BFE,W6
        MOV  		#VISI,W7
        NOP

        TBLRDL  	[W6--],[W7]
        NOP
        NOP
		NOPR

Command_set_Read_CW_Lo:
        TBLRDL  	[W6],[W7]
        NOP
        NOP
		NOPR

Command_set_Read_Flash_Init:
		NOP
		GOTO		0x200

        MOV  		#0x0000,W0
        MOV  		W0,TBLPAG
        MOV  		#0x0000,W6
        MOV  		#VISI,W7
        NOP

		NOPR
Command_set_Read_Flash_H:
		TBLRDH		[W6],[W7]
		NOP
		NOP
		NOPR
Command_set_Read_Flash_L:
		TBLRDL		[W6++],[W7]
		NOP
		NOP
		NOPR

Command_set_Erase_Init:
		NOP
		GOTO		0x200

        MOV  #0x404F,W10
        MOV  W10,NVMCON

		MOV		#0x0000,W0
		MOV		W0,TBLPAG
		MOV		#0x0000,W0
		NOP
		TBLWTL	W0,[W0]
		NOP
		NOP

        BSET  NVMCON, #15		;WR (BIT 15)
        NOP
        NOP
		NOPR

Command_set_PollWR:
        GOTO 0x200
        MOV  #VISI,W1
        MOV  NVMCON,W2
        MOV  W2,[W1]
        NOP
		NOPR

Command_set_Program_CW_Init:
		NOP
		GOTO		0x200

        MOV  #0x2BFC,W7

        MOV  #0x4003,W10
        MOV  W10,NVMCON

        MOV  #0x0000,W0
        MOV  W0,TBLPAG

        MOV  #0xFACC,W6				;(CW2)

		NOP
        TBLWTL  W6,[W7++]
        NOP
        NOP

        BSET  NVMCON, #15			;WR (BIT 15)
        NOP
        NOP
		NOPR


Command_set_Program_CW_Second:
        NOP
		GOTO 0x200

        MOV  #0x0000,W0
        MOV  W0,TBLPAG

        MOV  #0x3F7F,W6        ;(CW1)

        TBLWTL  W6,[W7++]
        NOP
        NOP

        BSET  NVMCON, #15		;WR (BIT 15)
        NOP
        NOP
		NOPR

Command_set_Program_flash_Init:
        NOP
		GOTO 0x200

        MOV  #0x4001,W10
        MOV  W10,NVMCON
		NOPR


Command_set_Program_flash_LoopA:

        TBLWTH.B  W2,[W7]
        NOP
        NOP
        TBLWTL  W1,[W7++]
        NOP
        NOP
		NOPR

Command_set_Program_flash_LoopB:
        BSET  NVMCON, #15				;WR (BIT 15)
        NOP
		NOP
		NOPR


;..............................................................................
; Code Section in Program Memory
;..............................................................................
;	Wnn registers global assignments
;   --------------------------------
;	W11: Address of the LATB register
;	W12: Mask of the currently selected MCU (1000 0000 0000 0000 to 0000 0001 0000 0000)
;	W13: Complement of the W12
;
;
;

.text                             ;Start of Code section
__reset:
        MOV #__SP_init, W15       ;Initalize the Stack Pointer
        MOV #0x0FFE, W0     	  ;Initialize the Stack Pointer Limit Register
        MOV W0, SPLIM
        NOP                       ;Add NOP to follow SPLIM initialization
        
        CALL _wreg_init           ;Call _wreg_init subroutine




        ;<<insert more user code here>>


;
; initialize
;
		MOV #0xFFFF,W1
		MOV W1,AD1PCFG			; turn off the Analog nature of Port A pins

		MOV	#0xFFFF, W1			; set PORTA as input.
		MOV	W1,TRISA

		CALL	Clear_all_RPx_outputs
		MOV	#0xFF00,W1			; set PortB 8-15 as "open drain" (PGD)
		MOV	W1,ODCB	
		MOV	#0x003F,W1			;set PortB 8-15 and PortB 6-7 as output (all PGD and MCLR, PGC)
		MOV	W1,TRISB
		MOV	#0xFFFF,W1			;set all PortB as open drain
		MOV		W1,LATB

		MOV		#0x8000,W12		;Set mask for MCU7
		COM		W12,W13			;Set the complement of the mask
		MOV		#LATB,W11		;Set the address of the PortB


		CALL 	Setup_UART


; main code


		ICSP_set_MCLR_1


		PrintStringFlash		idname

MainLoop:


		CALL	UART_Read_String
		strcmp		InBuf, cmd_Version
		BRA		NZ,ML1
;
; "Version" command
;
		CALL	UART_Print_String_prompt
		PrintStringFlash		idname

		BRA		MLx

ML1:
		strcmp		InBuf, cmd_Identify
		BRA		NZ,ML2
;
; "Identify" command
;


		CALL	UART_Print_String_prompt
		CALL 	IdentifyMCU
		CALL	UART_Print_String_CR

		BRA		MLx

ML2:
		strcmp		InBuf, cmd_ReadCW
		BRA		NZ,ML3
;
; "ReadCW" command
;

		CALL	UART_Print_String_prompt
		call ReadCW
		CALL	UART_Print_String_CR

		BRA		MLx

ML3:
		strcmp		InBuf, cmd_ReadFlash
		BRA		NZ,ML4
;
; "ReadFlash" command
;

		CALL	UART_Print_String_prompt
		call ICSP_ReadFlash
		CALL	UART_Print_String_CR

		BRA		MLx
ML4:
		strcmp		InBuf, cmd_ReadFlashBin
		BRA		NZ,ML5
;
; "ReadFlashBin" command
;

		call 	ICSP_ReadFlashBin
		CALL	UART_Print_String_prompt
		CALL	UART_Print_String_CR

		BRA		MLx

ML5:
		strcmp		InBuf, cmd_MCLR0
		BRA		NZ,ML6
;
; "MCLR0" command
;

		ICSP_set_MCLR_0

		CALL	UART_Print_String_prompt
		CALL	UART_Print_String_CR

		BRA		MLx

ML6:
		strcmp		InBuf, cmd_MCLR1
		BRA		NZ,ML7
;
; "MCLR1" command
;

		ICSP_set_MCLR_1

		CALL	UART_Print_String_prompt
		CALL	UART_Print_String_CR

		BRA		MLx

ML7:
		strcmp		InBuf, cmd_Erase
		BRA		NZ,ML8
;
; "Erase" command
;

		CALL 	ICSP_Erase

		CALL	UART_Print_String_prompt
		CALL	UART_Print_String_CR

		BRA		MLx

ML8:
		strcmp		InBuf, cmd_WriteCW
		BRA		NZ,ML9
;
; "WriteCW" command
;

		CALL 	ICSP_WriteCW

		CALL	UART_Print_String_prompt
		CALL	UART_Print_String_CR

		BRA		MLx

ML9:
		strcmp		InBuf, cmd_WriteBlock
		BRA		NZ,ML10
;
; "WriteBlock(nn)" command
;

		MOV		#BlockNumber,W0
		MOV		W2,[W0]				;Store Block number

		; Read data, store into RAM
		CALL	UART_Read_Block

		; Program Block
		CALL 	ICSP_ProgramBlock


		MOV		#BlockNumber,W0
		MOV		[W0],W2				;Read Block number
		MOV		#OutBuf,W0
		MOV		#0x00030,W1
		ADD.B	W1,W2,W2
		MOV.B	W2,[W0++]


		MOV		#0x0000,W2
		MOV.B	W2,[W0]

		CALL	UART_Print_String_prompt
		PrintStringRAM	OutBuf
		CALL	UART_Print_String_CR

		BRA		MLx

ML10:
		strcmp		InBuf, cmd_SelectMCU
		BRA		NZ,ML11
;
; "SelectMCU(nn)" command
;

		MOV		#CurrentMCU,W0
		MOV		W2,[W0]				;Store Selected MCU number

		MOV		#0x0100,W0
		SL		W0,W2,W12		;Set the mask for the current MCU
		COM		W12,W13			;Set the complement of the mask


		CALL	UART_Print_String_prompt
		CALL	UART_Print_String_CR

		BRA		MLx

ML11:
		CALL	UART_Print_String_prompt
		PrintStringFlash		Unknown_cmd


MLx:

				BRA		MainLoop


;---------------------------------------------------------------

done:
        BRA     done              ;Place holder for last line of executed code


;..............................................................................
; Subroutine ICSP_ProgramBlock
;..............................................................................
ICSP_ProgramBlock:
		PUSH		W0
		PUSH		W1
		PUSH		W2
		PUSH		W3
		PUSH		W4
		PUSH		W5
		PUSH		W6

		CALL		ICSP_enter
		ICSP_execute_code	Command_set_Program_flash_Init

		; Prepare the start address according to the block number
		MOV		#BlockNumber,W0
		MOV		[W0],W2				;Read Block number
		MOV		#(2*64*16),W1		;2= 2 words per instruction.
		MUL.UU	W1,W2,W4			;MUL 64*2, store in W4-W5 (W4=LSB), shift left 4 bits (*16)

		MOV		#0x0020,W2
		MOV		#0x0000,W1
		CALL	ICSP_send_opcodeP			;W2=MSB,W1=LSB ==> MOV  #0x0000,W0

		MOV		#0x0088,W2
		MOV		#0x0190,W1
		CALL	ICSP_send_opcodeP			;W2=MSB,W1=LSB ==> MOV  W0,TBLPAG

		MOV		#0xFFF0,W1
		AND		W1,W4,W4
		MOV		#0x000F,W1
		AND		W1,W5,W5
		MOV		#0x0007,W1
		IOR		W1,W4,W4
		MOV		#0x0020,W1
		IOR		W1,W5,W5
		MOV		W4,W1				;LSB
		MOV		W5,W2				;MSB
		CALL	ICSP_send_opcodeP			;W2=MSB,W1=LSB ==> MOV  #<DestAddress>,W7

		MOV		#BlockData,W0
		MOV		#64,W6

ICSP_WFLA:
		MOV.B	[W0++],W3
		MOV		#0x00FF,W2
		AND		W2,W3,W3
		SL		W3,#4,W3
		MOV		#0x0002,W2
		IOR		W2,W3,W1
		MOV		#0x0020,W2
		CALL	ICSP_send_opcodeP			;W2=MSB,W1=LSB ==> MOV  #<MSByte>,W2 = 200xx2

		MOV.B	[W0++],W4
		MOV.B	[W0++],W5
		LSR		W4,#4,W2
		MOV		#0x000F,W3
		AND		W3,W2,W2
		MOV		#0x0020,W3
		IOR		W3,W2,W2			;Set-up MSB

		SL		W5,#4,W1
		MOV		#0x0FF0,W3
		AND		W3,W1,W1
		SL		W4,#12,W4
		MOV		#0xF000,W3
		AND		W3,W4,W4
		IOR		W1,W4,W1

		MOV		#0x0001,W3
		IOR		W3,W1,W1			;Set-up LSW
		CALL	ICSP_send_opcodeP			;W2=MSB,W1=LSB ==> MOV  #<LSWord>,W1 = 2xxxx1

		ICSP_execute_code	Command_set_Program_flash_LoopA
		
		DEC		W6,W6
		BRA		NZ,ICSP_WFLA

		ICSP_execute_code	Command_set_Program_flash_LoopB

		CALL	ICSP_Poll_till_WR_cleared


		ICSP_execute_code	Command_set_Reset


		ICSP_set_MCLR_0

		POP		W6
		POP		W5
		POP		W4
		POP		W3
		POP		W2
		POP		W1
		POP		W0


		RETURN
;..............................................................................
; Subroutine ICSP_WriteCW
;..............................................................................
ICSP_WriteCW:

		PUSH		W0
		PUSH		W1
		PUSH		W2

		CALL		ICSP_enter

		ICSP_execute_code	Command_set_Program_CW_Init

		CALL	ICSP_Poll_till_WR_cleared

		ICSP_execute_code	Command_set_Program_CW_Second

		CALL	ICSP_Poll_till_WR_cleared


		ICSP_set_MCLR_0

		POP		W2
		POP		W1
		POP		W0

		RETURN
;..............................................................................
; Subroutine ICSP_Erase
;..............................................................................
ICSP_Erase:

		PUSH		W0
		PUSH		W1
		PUSH		W2

		CALL		ICSP_enter

		ICSP_execute_code	Command_set_Erase_Init

		CALL	ICSP_Poll_till_WR_cleared


		ICSP_set_MCLR_0

		POP		W2
		POP		W1
		POP		W0

		RETURN
;..............................................................................
; Subroutine ICSP_Poll_till_WR_cleared
;..............................................................................
ICSP_Poll_till_WR_cleared:

ICSP_Poll:
		ICSP_execute_code	Command_set_PollWR
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		NOP
		MOV		[W0],W1

		MOV		#0x8000,W0
		AND		W0,W1,W1
		BRA		NZ,ICSP_Poll

		RETURN
;..............................................................................
; Subroutine ICSP_enter
;..............................................................................
ICSP_enter:

		PUSH		W0
		PUSH		W1
		PUSH		W2

		ICSP_set_PGD_0
		ICSP_set_PGC_0
		ICSP_set_MCLR_0
		NOP
		ICSP_set_MCLR_1
		ICSP_set_MCLR_0
;
; send key
;

		MOV		#0xC2B2,W1
		MOV		#0x8A12,W2			; W2:W1 = the reverse of 4D43-4851
		MOV		#0x0020,W0
		CALL	ICSP_send_data

		CALL	wait_p19

		ICSP_set_MCLR_1

		wait_p7

		MOV		#0x0000,W2
		MOV		#0x0000,W1
		MOV		#5,W0
		CALL	ICSP_send_data		;send 5 zeroes

		ICSP_execute_code	Command_set_Reset

		POP		W2
		POP		W1
		POP		W0

		RETURN
;..............................................................................
; Subroutine ICSP_ReadFlashBin
;..............................................................................
ICSP_ReadFlashBin:
		PUSH		W0
		PUSH		W1
		PUSH		W2

		CALL		ICSP_enter

		ICSP_execute_code	Command_set_Read_Flash_Init

		MOV			#88,W4				;number of blocks
ICSP_RFb1:

		MOV		#64,W5					;number of instructions per block

ICSP_RFb2:

		ICSP_execute_code	Command_set_Read_Flash_H
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		MOV		[W0],W1
		MOV		#OutBuf,W0
		MOV.B	W1,[W0++]
		MOV		W0,OutBuf_index


		ICSP_execute_code	Command_set_Read_Flash_L
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		MOV		[W0],W1
		MOV		OutBuf_index,W0
		MOV		#0xFF00,W2
		AND		W1,W2,W2
		LSR		W2,#8,W2
		MOV.B	W2,[W0++]
		MOV		#0x00FF,W2
		AND		W1,W2,W2
		MOV.B	W2,[W0]

		CALL	UART_send_3bytes


		DEC		W5,W5
		BRA		NZ,ICSP_RFb2

		ICSP_execute_code	Command_set_Reset
;		PrintStringFlash		block
		

		DEC		W4,W4
		BRA		NZ,ICSP_RFb1

		ICSP_set_MCLR_0

		POP		W2
		POP		W1
		POP		W0



		RETURN
;..............................................................................
; Subroutine ICSP_ReadFlash
;..............................................................................
ICSP_ReadFlash:
		PUSH		W0
		PUSH		W1
		PUSH		W2

		CALL		ICSP_enter

		ICSP_execute_code	Command_set_Read_Flash_Init

		MOV			#88,W4				;number of blocks
ICSP_RF1:

		MOV		#64,W5					;number of instructions per block

ICSP_RF2:


		ICSP_execute_code	Command_set_Read_Flash_H
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		MOV		[W0],W1
		CALL	OutBuf_write_W1_hex
		PrintStringRAM	OutBuf


		ICSP_execute_code	Command_set_Read_Flash_L
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		MOV		[W0],W1
		CALL	OutBuf_write_W1_hex
		PrintStringRAM	OutBuf

		PrintStringFlash		dash		

		DEC		W5,W5
		BRA		NZ,ICSP_RF2

		ICSP_execute_code	Command_set_Reset
		PrintStringFlash		block
		

		DEC		W4,W4
		BRA		NZ,ICSP_RF1

		ICSP_set_MCLR_0

		POP		W2
		POP		W1
		POP		W0



		RETURN
;..............................................................................
; Subroutine ReadCW
;..............................................................................
ReadCW:
		PUSH		W0
		PUSH		W1
		PUSH		W2

		CALL		ICSP_enter

		ICSP_execute_code	Command_set_Read_CW_Hi
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		MOV		[W0],W1
		CALL	OutBuf_write_W1_hex
		PrintStringRAM	OutBuf
		PrintStringFlash		dash		


		ICSP_execute_code	Command_set_Read_CW_Lo
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		MOV		[W0],W1
		CALL	OutBuf_write_W1_hex
		PrintStringRAM	OutBuf
		

		ICSP_set_MCLR_0

		POP		W2
		POP		W1
		POP		W0


		RETURN
;..............................................................................
; Subroutine Identify MCU
;..............................................................................
IdentifyMCU:

		PUSH		W0
		PUSH		W1
		PUSH		W2

		CALL		ICSP_enter

		ICSP_execute_code	Command_set_Read_ID_Hi
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		MOV		[W0],W1

		CALL	OutBuf_write_W1_hex
		PrintStringRAM	OutBuf


		ICSP_execute_code	Command_set_Read_ID_Lo
        CALL	ICSP_receive_word

		MOV		#VISIout,W0
		MOV		[W0],W1

		CALL	OutBuf_write_W1_hex
		PrintStringRAM	OutBuf

		ICSP_set_MCLR_0

		POP		W2
		POP		W1
		POP		W0

		RETURN
;..............................................................................
;	Subroutine ICSP_send_code
;
;	This subroutine sends for execution a batch of code.
;	The code is stored into the flash memory and ends with the NOPR command
;	(the NOPR command is not transmited).
;..............................................................................
ICSP_send_code:
		PUSH		W1
		PUSH		W2
		PUSH		W3
		PUSH		W4

ICSP_SC:
		MOV			#0x00FF,W4
		MOV			#0x0000,W2
		TBLRDH.B	[W3],W2
		TBLRDL		[W3++],W1
		CP			W4,W2
		BRA			NZ,ICSP_SC2
		CP0			W1
		BRA			NZ,ICSP_SC2
		BRA			ICSP_SC3
ICSP_SC2:
		CALL	ICSP_send_opcodeP
		BRA			ICSP_SC

ICSP_SC3:

		POP		W4
		POP		W3
		POP		W2
		POP		W1
		RETURN
;..............................................................................
; Subroutine OutBuf_write_W1_bin
;	The input data is held in W1. The data are destroyed.
;	The data are transmited in binary format, MSB first (2 bytes)
;..............................................................................
OutBuf_write_W1_bin:

		PUSH	W0
		PUSH	W2

		MOV		#OutBuf,W0

		MOV		#0xFF00,W2
		AND		W1,W2,W2
		LSR		W2,#8,W2
		MOV.B	W2,[W0++]

		MOV		#0x00FF,W2
		AND		W1,W2,W2
		MOV.B	W2,[W0++]

		MOV		#0x0000,W2
		MOV.B	W2,[W0++]


		POP		W2
		POP		W0

		RETURN
;..............................................................................
; Subroutine OutBuf_write_W1_hex
;	The input data is held in W1. The data are destroyed.
;..............................................................................
OutBuf_write_W1_hex:

		PUSH	W0
		PUSH	W2
		PUSH	W3

		MOV		#OutBuf,W0

		MOV		#0xF000,W2
		AND		W1,W2,W2
		LSR		W2,#12,W2
		MOV		#0x000A,W3
		CPSLT	W2,W3
		ADD		#7,W2
		ADD		#0x30,W2
		MOV.B	W2,[W0++]

		MOV		#0x0F00,W2
		AND		W1,W2,W2
		LSR		W2,#8,W2
		CPSLT	W2,W3
		ADD		#7,W2
		ADD		#0x30,W2
		MOV.B	W2,[W0++]

		MOV		#0x00F0,W2
		AND		W1,W2,W2
		LSR		W2,#4,W2
		CPSLT	W2,W3
		ADD		#7,W2
		ADD		#0x30,W2
		MOV.B	W2,[W0++]

		MOV		#0x000F,W2
		AND		W1,W2,W2
		CPSLT	W2,W3
		ADD		#7,W2
		ADD		#0x30,W2
		MOV.B	W2,[W0++]

		MOV		#0x0000,W2
		MOV.B	W2,[W0++]


		POP		W3
		POP		W2
		POP		W0

		RETURN
;..............................................................................
; Subroutine ICSP_receive_word
;	This subroutine receives a word through the VISI register
;	The SIX command 1000 is transmited first
;	It leaves PGC,PGD in state 0.
;	Returns the result into RAM location "VISIout"
;	Uses ICSP_send_data, ICSP_send_opcodeP
;..............................................................................
ICSP_receive_word:

		MOV		#0x0000,W2
		MOV		#0x0FF1,W1
		MOV		#12,W0
		CALL	ICSP_send_data		;send SIX command 1000 plus 8 zeroes

		MOV		#0x0000,W1
		ICSP_set_PGC_0
		ICSP_set_PGD_1			;Set PGD =1 ==> open drain

		MOV		#16,W0
ICSP_rw1:

		ICSP_set_PGC_1
		NOP						;small delay

		MOV		PORTB,W2		;Read PGD register

		BCLR	SR,#0x00		;Clear Carry
		AND		W12,W2,W2		;Mask the current MCU's PGD
		BRA		Z,ICSP_rw2
		BSET	SR,#0x00		;Set Carry
ICSP_rw2:
		RRC		W1,W1	

		ICSP_set_PGC_0
		NOP						;small delay


		DEC		W0,W0
		BRA		NZ,ICSP_rw1

		ICSP_set_PGC_0
		NOP						;small delay
		ICSP_set_PGD_0

		MOV		#VISIout,W0
		MOV		W1,[W0]

		;Send opcode 000000 (NOP)
		MOV		#0x0000,W2
		MOV		#0x0000,W1
		CALL	ICSP_send_opcodeP

		RETURN
;..............................................................................
; Subroutine ICSP_send_opcodeP
;	This subroutine transmits the opcode stored into registers W2 (MSB) - W1 (LSB)
;	The data are transmited from LSB to MSB.
;	The SIX command 0000 is transmited first
;	It leaves PGC,PGD in state 0.
;	It destroys W0, W1, W2
;	Uses ICSP_send_data
;..............................................................................
ICSP_send_opcodeP:
		PUSH	W0
		PUSH	W1
		PUSH	W2


		PUSH	W1
		PUSH	W2

		MOV		#0x0000,W2
		MOV		#0x0000,W1
		MOV		#4,W0
		CALL	ICSP_send_data		;send SIX command 0000

		POP		W2
		POP		W1

		MOV		#24,W0
		CALL	ICSP_send_data		;send opcode


		POP		W2
		POP		W1
		POP		W0
		RETURN
;..............................................................................
; Subroutine ICSP_send_data
;	This subroutine transmits the data stored into registers W2 (MSB) - W1 (LSB)
;	The data are transmited from LSB to MSB.
;	Register W0 keeps the number of bits to be trasnmited.
;	It leaves PGC,PGD in state 0.
;	It destroys W0, W1, W2
;	Uses ICSP_clock
;..............................................................................
ICSP_send_data:

		ICSP_set_PGC_0

ICSP_sd1:
		ICSP_set_PGD_0
		BTSC	W1,#0		;test bit0 of W1
		ICSP_set_PGD_1

		CALL	ICSP_clock

		RRC		W2,W2
		RRC		W1,W1
		DEC		W0,W0
		BRA		NZ,ICSP_sd1

		ICSP_set_PGC_0
		ICSP_set_PGD_0

		RETURN
;..............................................................................
; Subroutine ICSP_clock
;	This subroutine generates a positive pusle in PGC (PortB-7).
;	It leaves PGC in state 0.
;
;	It uses the masks and dtata stored into registers W11, W12, W13
;	Does not modify any register (except PortB-7 bit)
;..............................................................................
ICSP_clock:

		ICSP_set_PGC_0

		ICSP_set_PGC_1

		ICSP_set_PGC_0

		RETURN
;..............................................................................
; Subroutine stringcompare
;	This subroutine compares two strings.
;	The first string is stored into RAM. (indexed by W0)
;	The second string is stored into flash.
;	There is a special care when the flash-based string contains
;	the non-printable character 0x01 as a stop byte
;	(Note: for safety reasons the flash-written string must 
;	also contain the 0x00 termination byte.)
;..............................................................................
stringcompare:


		PUSH	W3
		PUSH	W4
		PUSH	W5

		MOV	#0x0000,W3			;flag
		MOV	#0x0000,W4			;zero

stringcompareloop:
		TBLRDL.B	[W1++],W2
		MOV.B		[W0],W5
		CP.B	W2,[W0]
		BRA		NZ,stringcompare2		;difference found
		CP.B	W4,[W0++]
		BRA		Z,stringcompare3		;end-of-string
		BRA		stringcompareloop

stringcompare2:
		MOV		#0x0001,W4
		CP.B	W4,W2
		BRA		NZ,stringcompare4

		; string matches till the flash-written string keeps 0x01
		; This means that the rest of the input is a number !!!!

		MOV.B		[W0++],W3
		MOV			#0x000F,W4
		AND			W3,W4,W4
		MUL.UU		W4,#10,W4			;Attention: W5 is afected as well !!!
		MOV			W4,W2

		MOV.B		[W0],W3
		MOV			#0x000F,W4
		AND			W3,W4,W4
		ADD			W4,W2,W2

		; The Integer parameter is stored into W2

		MOV	#0x0000,W3			;flag
		MOV	#0x0000,W4			;zero
		BRA		stringcompare3

stringcompare4:
		MOV		#0x0001,W3		;not equal flag
		MOV		#0x0000,W4
stringcompare3:
		CP		W3,W4

		POP	W5
		POP	W4
		POP	W3

		RETURN
;..............................................................................
; Subroutine UART_Read_Block
;	This Procedure reads 64 * 3 bytes from the UART.
;	The received string is written in buffer that starts at
;	address "BlockData" of RAM. 
;..............................................................................

UART_Read_Block:

		PUSH	W0
		PUSH	W1
		PUSH	W2

		MOV		#BlockData,W0
		MOV		#(64*3),W2

RBpoll:
		BTST	U1STA,#0	;wait till Receive Interrupt is set.
		BRA		Z,RBpoll

		MOV		U1RXREG,W1			;Read UART
		MOV.B	W1,[W0++]

		DEC		W2,W2
		BRA		NZ,RBpoll		;Next


		POP	W2
		POP	W1
		POP	W0

		RETURN
;..............................................................................
; Subroutine UART_Read_String
;	This Procedure reads a string from the UART.
;	It remains in a loop till a CR character is received.
;   It ignores the LF characters
;	The received string is written in buffer that starts at
;	address "InBuf" of RAM. The string ends with a 00h. The CR or LF
;	is NOT written to the string!!
;..............................................................................

UART_Read_String:

		PUSH	W0
		PUSH	W1
		PUSH	W2

		MOV		#InBuf,W0

RSpoll:
		BTST	U1STA,#0	;wait till Receive Interrupt is set.
		BRA		Z,RSpoll

		MOV		U1RXREG,W1			;Read UART

		MOV		#0x000A,W2	
		CP		W1,W2
		BRA		Z,RSpoll		;Ignore LF

		MOV		#0x000D,W2	
		CP		W1,W2
		BRA		Z,RS2

							; non CR character
		MOV.B		W1,[W0++]
		BRA		RSpoll

RS2:						; CR 
		MOV		#0x0000,W1
		MOV.B	W1,[W0]

		POP	W2
		POP	W1
		POP	W0

		RETURN
;..............................................................................
;Subroutine: UART_send_3bytes
; This subroutine transmits 3 bytes using UART.
; The 3 bytes are supposed to be in the RAM memory. 
;..............................................................................
UART_send_3bytes:

		PUSH	W1
		PUSH	W2

		MOV		#OutBuf,W0
		MOV		#3,W2

S3Bloop:
		MOV.B	[W0++],W1

		MOV	W1,U1TXREG
S3Bpoll:
		BTST	U1STA,#0x08
		BRA		Z,S3Bpoll

		DEC		W2,W2
		BRA		NZ,S3Bloop



		POP	W2
		POP	W1

		RETURN

;..............................................................................
;Subroutine: UART_Print_string_RAM
; This subroutine transmits a string using UART.
; The string is supposed to be in the RAM memory. 
; Prior to the calling the program must store into W0 the 
; addresse's offset. (NOTE: W0 is modified)
; The string ends with character 00 hex.
;..............................................................................
UART_Print_String_RAM:

		PUSH	W1
		PUSH	W2

		MOV		#0x0000,W2

PSRloop:
		MOV.B	[W0++],W1
		CP.B	W1,W2
		BRA		Z,PSRexit

		MOV	W1,U1TXREG
PSRpoll:
		BTST	U1STA,#0x08
		BRA		Z,PSRpoll

		BRA		PSRloop


PSRexit:

		POP	W2
		POP	W1

		RETURN

;..............................................................................
; Subroutine:	UART_Print_String_CR
;..............................................................................
UART_Print_String_CR:
		PUSH	W0

		MOV	#tblpage(CR),W0
		MOV	W0,_TBLPAG
		MOV	#tbloffset(CR),W0

		CALL	UART_Print_String_flash

		POP		W0

		RETURN
;..............................................................................
; Subroutine:	UART_Print_String_prompt
;..............................................................................
UART_Print_String_prompt:
		PUSH	W0

		MOV	#tblpage(prompt),W0
		MOV	W0,_TBLPAG
		MOV	#tbloffset(prompt),W0

		CALL	UART_Print_String_flash

		POP		W0

		RETURN
;..............................................................................
;Subroutine: UART_Print_string_flash
; This subroutine transmits a string using UART.
; The string is supposed to be in the program memory. 
; Prior to the calling the program must set the TBLPAG reg, and store into W0 the 
; addresse's offset. (NOTE: W0 is modified)
; The string ends with character 00 hex.
;..............................................................................
UART_Print_String_flash:

		PUSH	W1
		PUSH	W2

		MOV		#0x0000,W2

PSloop:
		TBLRDL.B	[W0++],W1
		CP.B	W1,W2
		BRA		Z,PSexit

		MOV	W1,U1TXREG
PSpoll:
		BTST	U1STA,#0x08
		BRA		Z,PSpoll

		BRA		PSloop


PSexit:

		POP	W2
		POP	W1

		RETURN

;..............................................................................
;Subroutine: Setup UART1 for communication with the PC
;..............................................................................

Setup_UART:

		PUSH	W1

		MOV	#0x1F03,W1
		MOV	W1,RPINR18			; RxD --> RP3

		MOV	#0x0003,W1
		MOV	W1,RPOR1			; TxD --> RP2

		MOV	#103,W1
		MOV	W1,U1BRG			; BRG <-- 103(dec) (baud rate = 9600 @ 8MHZ with BRGH=1)
		MOV	#0x0008,W1			; 0000 0000 0000 1000
		MOV	W1,U1MODE			; setup UART: UARTEN=0, RXINV=0, BRGH=1, 8-bit, no parity, 1 stop.
		MOV	#0x8008,W1
		MOV	W1,U1MODE			; Enable UART
		MOV	#0x04C0,W1
		MOV	W1,U1STA			; setup UART


		MOV	#0x0008,W1
		MOV	W1,U1MODE
		MOV	#0x8008,W1
		MOV	W1,U1MODE
		MOV	#0x04C0,W1
		MOV	W1,U1STA			; setup UART

		POP	W1
		RETURN
;..............................................................................
; Subroutine:	Wait_p19
;
; Wait for 1 msec
;
; 1msec = 16000 ops @ 32MHZ = 500 * XTAL
; NOP,DEC,BRA (taken) = 4 ops ==> 1msec <= (130 * XTAL)  loops
;..............................................................................
wait_p19:
		PUSH	W8
		NOP

		MOV		#(130*XTAL),W8
w_p19:
		NOP
		DEC	W8,W8
		BRA	NZ,w_p19

		POP	W8

		RETURN
;..............................................................................
;Subroutine: Clear RP2-RP15 pins Output (==> set all RPORx to 0000)
;..............................................................................
Clear_all_RPx_outputs:

		PUSH	W1

		MOV		#0x0000,W1
		MOV		W1,RPOR1		; clear RP3, RP2
		MOV		W1,RPOR2		; clear RP5, RP4
		MOV		W1,RPOR3		; clear RP7, RP6
		MOV		W1,RPOR4		; clear RP9, RP8
		MOV		W1,RPOR5		; clear RP11, RP10
		MOV		W1,RPOR6		; clear RP13, RP12
		MOV		W1,RPOR7		; clear RP15, RP14

		POP		W1
		RETURN
;..............................................................................
;Subroutine: Initialization of W registers to 0x0000
;..............................................................................

_wreg_init:
        CLR W0
        MOV W0, W14
        REPEAT #12
        MOV W0, [++W14]
        CLR W14
        RETURN



;..............................................................................
;Timer 1 Interrupt Service Routine
;Example context save/restore in the ISR performed using PUSH.D/POP.D
;instruction. The instruction pushes two words W4 and W5 on to the stack on
;entry into ISR and pops the two words back into W4 and W5 on exit from the ISR
;..............................................................................

;__T1Interrupt:
;        PUSH.D W4                  ;Save context using double-word PUSH

        ;<<insert more user code here>>

;        BCLR IFS0, #T1IF           ;Clear the Timer1 Interrupt flag Status
                                   ;bit.
;        POP.D W4                   ;Retrieve context POP-ping from Stack
;        RETFIE                     ;Return from Interrupt Service routine



;--------End of All Code Sections ---------------------------------------------

.end                               ;End of program code in this file



